import { Shadow, HTML, SVG, API, El, shadow, html } from '../../index';
import { Coroutine } from 'spica/coroutine';
import { Sequence } from 'spica/sequence';
import { Attrs, Children } from '../../internal';

declare global {
  interface ShadowHostElementTagNameMap {
    'custom-tag': HTMLElement;
  }
  interface HTMLElementTagNameMap {
    'custom': HTMLElement;
  }
}

declare const _: { shuffle<T>(as: T[]): T[]; };

describe('Integration: Typed DOM', function () {
  describe('spec', function () {
    it('call', function () {
      assert(HTML('p').element.outerHTML === '<p></p>');
      assert(HTML('p', 'a').element.outerHTML === '<p>a</p>');
      assert(HTML('p', undefined, 'a').element.outerHTML === '<p>a</p>');
      assert(HTML('p', { class: 'class' }, 'a', (h, t) => h(t, { id: 'id' })).element.outerHTML === '<p id="id" class="class">a</p>');
    });

    it('html', function () {
      assert(HTML.html().element instanceof HTMLElement);
      assert(HTML.html().element.outerHTML === '<html></html>');
    });

    it('svg', function () {
      assert(SVG.svg().element instanceof SVGElement);
      assert(SVG.svg().element.outerHTML === '<svg></svg>');
    });

    it('empty', function () {
      const dom = HTML.p();
      assert(dom.element.outerHTML === '<p></p>');
      assert(dom.children === undefined);
    });

    it('factory', function () {
      const dom = HTML.p({ class: 'test' }, [HTML.a()], (h, tag, _, children) => {
        assert(children.every(child => child.element instanceof HTMLAnchorElement));
        return html('div').appendChild(h(tag, { id: 'test' }));
      });
      assert(dom.element.id === 'test');
      assert(dom.element.className === 'test');
      assert(dom.children.length === 1);
      assert(dom.children.every(child => child.element instanceof HTMLAnchorElement));
    });

    it('text', function () {
      const dom = HTML.p(`a`);
      assert(dom.element.outerHTML === '<p>a</p>');
      assert(dom.children === 'a');
    });

    it('text children update', function () {
      const dom = HTML.p(`a`);
      dom.children = 'b';
      assert(dom.element.outerHTML === '<p>b</p>');
      assert(dom.children === 'b');
    });

    it('text with factory', function () {
      const dom = HTML.p(`a`, (h, tag) =>
        h(tag, { id: 'test' }));
      assert(dom.element.id === 'test');
      assert(dom.children === 'a');
    });

    it('collection', function () {
      const dom = HTML.ul([
        HTML.li(`1`),
        HTML.li(`2`)
      ]);
      assert(dom.element.outerHTML === '<ul><li>1</li><li>2</li></ul>');
      assert(dom.children.length === 2);
      assert(dom.children.every(({ element }, i) => element === dom.element.children[i]));
    });

    it('collection children update', function () {
      this.timeout(9 * 1e3);

      const dom = HTML.ul([
        HTML.li(`1`)
      ]);
      assert.doesNotThrow(() => dom.children = dom.children);
      assert.throws(() => dom.children = HTML.ul([HTML.li(`1`)]).children);
      dom.children = [
        HTML.li('2'),
        HTML.li('3')
      ];
      assert(dom.element.outerHTML === '<ul><li>2</li><li>3</li></ul>');
      assert(dom.children.length === 2);
      assert(dom.children.every(({ element }, i) => element === dom.element.children[i]));
      dom.children = [
        HTML.li('4')
      ];
      assert(dom.element.outerHTML === '<ul><li>4</li></ul>');
      assert(dom.children.length === 1);
      assert(dom.children.every(({ element }, i) => element === dom.element.children[i]));

      // Property test
      const el = HTML.ul([HTML.li('')]);
      const es = Sequence.from([HTML.li('1'), HTML.li('2'), HTML.li('3')]);
      el.children = [];
      Sequence.from([
        ...es.permutations(),
        ...es.permutations().bind(es => Sequence.from(es).subsequences()),
      ])
        .extract()
        .forEach(es => {
          el.children = es;
          assert.deepStrictEqual(el.children, es);
        });
    });

    it('collection children partial update', function () {
      const dom = HTML.ul([
        HTML.li()
      ] as const);
      //assert.throws(() => dom.children[0] = dom.children[0]);
      //assert.throws(() => dom.children[0] = HTML.li());
      //assert.throws(() => dom.children.push(HTML.li()));
      //assert.throws(() => dom.children.pop());
      //assert.throws(() => dom.children.length = 0);
      assert(dom.children.length === 1);
      assert(dom.children.every(({ element }, i) => element === dom.element.children[i]));
    });

    it('collection with factory', function () {
      const dom = HTML.ul([], (h, tag) =>
        h(tag, { id: 'test' }));
      assert(dom.element.id === 'test');
      assert.deepStrictEqual(dom.children, []);
    });

    it('struct', function () {
      const dom = HTML.article({
        title: HTML.h1(`title`),
        content: HTML.p([HTML.a()])
      });
      assert(dom.element.outerHTML === '<article><h1>title</h1><p><a></a></p></article>');
      assert(dom.children.title.element === dom.element.firstChild);
      assert(dom.children.content.element === dom.element.lastChild);
    });

    it('struct empty', function () {
      const dom = HTML.div({});
      assert.deepStrictEqual(dom.children, {});
    });

    it('struct children update', function () {
      const dom = HTML.article({
        title: HTML.h1(`a`)
      });
      assert.doesNotThrow(() => dom.children = dom.children);
      assert.throws(() => dom.children = HTML.article({ title: HTML.h1(`b`) }).children);
      assert(dom.children.title.element === dom.element.firstChild);
      assert(dom.children.title.element.textContent === 'a');
      assert(dom.children.title.children === 'a');
      dom.children = {
        title: HTML.h1(`b`)
      };
      assert(dom.children.title.element === dom.element.firstChild);
      assert(dom.children.title.element.textContent === 'b');
      assert(dom.children.title.children === 'b');
    });

    it('struct children partial update', function () {
      const dom = HTML.article({
        title: HTML.h1(`a`)
      });
      assert.doesNotThrow(() => dom.children.title = dom.children.title);
      assert.throws(() => dom.children.title = HTML.article({ title: HTML.h1(`b`) }).children.title);
      assert(dom.children.title.element === dom.element.firstChild);
      assert(dom.children.title.element.textContent === 'a');
      assert(dom.children.title.children === 'a');
      dom.children.title = HTML.h1(`b`);
      assert(dom.children.title.element === dom.element.firstChild);
      assert(dom.children.title.element.textContent === 'b');
      assert(dom.children.title.children === 'b');
      dom.children.title.children = 'c';
      assert(dom.children.title.element === dom.element.firstChild);
      assert(dom.children.title.element.textContent === 'c');
      assert(dom.children.title.children === 'c');
    });

    it('struct with factory', function () {
      const dom = HTML.article({}, (h, tag) =>
        h(tag, { id: 'test' }));
      assert(dom.element.id === 'test');
      assert.deepStrictEqual(dom.children, {});
    });

    it('attr', function () {
      const dom = HTML.div({ id: 'test', class: 'test' });
      assert(dom.element.id === 'test');
      assert(dom.element.getAttribute('class') === 'test');
      assert(dom.children === undefined);
    });

    it('attr with factory', function () {
      const dom = HTML.div({ id: 'test' }, (h, tag) =>
        h(tag, { id: 'id', class: 'test' }));
      assert(dom.element.id === 'test');
      assert(dom.element.className === 'test');
      assert(dom.children === undefined);
    });

    it('attr with text', function () {
      const dom = HTML.div({ id: 'test', class: 'test' }, '');
      assert(dom.element.id === 'test');
      assert(dom.element.getAttribute('class') === 'test');
      assert(dom.children === '');
    });

    it('attr with collection', function () {
      const dom = HTML.div({ id: 'test', class: 'test' }, []);
      assert(dom.element.id === 'test');
      assert(dom.element.getAttribute('class') === 'test');
      assert.deepStrictEqual(dom.children, []);
    });

    it('attr with struct', function () {
      const dom = HTML.div({ id: 'test', class: 'test' }, {});
      assert(dom.element.id === 'test');
      assert(dom.element.getAttribute('class') === 'test');
      assert.deepStrictEqual(dom.children, {});
    });

    it('attr with text and factory', function () {
      const dom = HTML.div({ id: 'test' }, '', (h, tag) =>
        h(tag, { id: 'id', class: 'test' }));
      assert(dom.element.id === 'test');
      assert(dom.element.className === 'test');
      assert(dom.children === '');
    });

    it('attr with collection and factory', function () {
      const dom = HTML.div({ id: 'test' }, [], (h, tag) =>
        h(tag, { id: 'id', class: 'test' }));
      assert(dom.element.id === 'test');
      assert(dom.element.className === 'test');
      assert.deepStrictEqual(dom.children, []);
    });

    it('attr with struct and factory', function () {
      const dom = HTML.div({ id: 'test' }, {}, (h, tag) =>
        h(tag, { id: 'id', class: 'test' }));
      assert(dom.element.id === 'test');
      assert(dom.element.className === 'test');
      assert.deepStrictEqual(dom.children, {});
    });

    it('listen', function (done) {
      HTML.a({ onclick: ev => void assert(ev instanceof Event) ?? done() }).element.click();
    });

    it('sanitize', function () {
      const dom = HTML.div('<script>');
      assert(dom.element.innerHTML === '&lt;script&gt;');
      assert(dom.children === '<script>');
      dom.children = '<script>';
      assert(dom.element.innerHTML === '&lt;script&gt;');
      assert(dom.children === '<script>');
    });

    it('scope', function () {
      const template = `$scope {}\n  $scope {}`;
      const result = template.replace(/\$scope/g, '#test');
      assert(HTML.div({ id: 'test' }, [HTML.style(template)]).children[0].element.innerHTML === result);
      assert(HTML.div({ id: 'test' }, { style: HTML.style(template) }).children.style.element.innerHTML === result);
      assert(HTML.div({ id: 'test' }, [HTML.style(`<script>`)]).children[0].element.children.length === 0);
      console.debug('id', HTML.div([HTML.style(template)]).element.className);
      assert(HTML.div([HTML.style(template)]).element.className.startsWith('rnd-'));
      assert(HTML.div([HTML.style(template)]).children[0].element.innerHTML.match(/\.[\w\-]+\s/gm)!.length === 2);
      assert(Shadow.div([HTML.style(template)]).element.className === '');
      assert(Shadow.div([HTML.style(template)]).children[0].element.innerHTML === template.replace(/\$scope/g, ':host'));
    });

    it('clear', function () {
      assert(HTML.p(() => HTML.p('a').element).element.innerHTML === 'a');
      assert(HTML.p(() => HTML.p('a').element).children === undefined);
      assert(HTML.p('', () => HTML.p('a').element).element.innerHTML === '');
      assert(HTML.p('', () => HTML.p('a').element).children === '');
      assert(HTML.p([], () => HTML.p('a').element).element.childNodes.length === 0);
      assert.deepStrictEqual(HTML.p([], () => HTML.p('a').element).children, []);
      assert(HTML.p({}, () => HTML.p('a').element).element.childNodes.length === 0);
      assert.deepStrictEqual(HTML.p({}, () => HTML.p('a').element).children, {});
    });

    it('fragment', function () {
      HTML.div([HTML.p(() => document.createDocumentFragment().appendChild(html('p')))]);
    });

    it('parameter combination', function () {
      Sequence.from([
        [{ id: 'id' }],
        [undefined, '', [], {}],
        [() => html('div')]] as const)
        .mapM(v => Sequence.from(v))
        .bind(v => Sequence.from(v).filterM(() => Sequence.from([false, true])))
        .extract()
        .forEach(params => {
          HTML.div(...params as any);
        });
    });

    it('extend', function () {
      window.customElements.define('custom-tag', class extends HTMLElement { });
      assert(Shadow('custom-tag').element.outerHTML === '<custom-tag></custom-tag>');
      assert(HTML('custom-tag').element.outerHTML === '<custom-tag></custom-tag>');
      assert(HTML.custom().element.outerHTML === '<custom></custom>');
    });

    it('swap', function () {
      const el = HTML.article([HTML.p()]);
      const children = el.children;
      assert.throws(() => HTML.article(children));
      el.children = [HTML.p()];
      assert(HTML.article(children));
    });

    it('observe text', function () {
      const el = HTML.span(
        {
          onmutate: (ev, el = ev.target as HTMLElement) =>
            el.textContent += el.textContent!,
        },
        'a');
      assert(el.children === 'aa');
      el.children = 'b';
      assert(el.children === 'bb');
    });

    it('observe collection', function () {
      const listeners: Record<string, EventListener> = {
        onconnect: (ev, el = ev.target as HTMLElement) =>
          el.textContent += el.textContent!.toUpperCase(),
        ondisconnect: (ev, el = ev.target as HTMLElement) =>
          el.textContent += el.textContent!,
      };
      const el = HTML.ul([
        HTML.li(listeners, 'a'),
        HTML.li(listeners, 'b'),
      ]);
      assert.deepStrictEqual(
        el.children.map(el => el.children),
        [
          'aA',
          'bB',
        ]);
      el.children = [
        el.children[1],
        HTML.li(listeners, 'c'),
      ];
      assert.deepStrictEqual(
        el.children.map(el => el.children),
        [
          'bB',
          'cC',
        ]);
      assert.deepStrictEqual(
        el.children.map(v => v.element),
        [...el.element.children]);
    });

    it('observe record', function () {
      const listeners: Record<string, EventListener> = {
        onconnect: (ev, el = ev.target as HTMLElement) =>
          el.textContent += el.textContent![0].toUpperCase(),
        ondisconnect: (ev, el = ev.target as HTMLElement) =>
          el.textContent += el.textContent![0].toLowerCase(),
      };
      const el = HTML.ul({
        a: HTML.li(listeners, 'a'),
        b: HTML.li(listeners, 'b'),
        c: HTML.li(listeners, 'c'),
        d: HTML.li(listeners, 'd'),
        e: HTML.li(listeners, 'e'),
      });
      assert.deepStrictEqual(
        Object.entries(el.children).map(([k, v]) => [k, v.children]),
        [
          ['a', 'aA'],
          ['b', 'bB'],
          ['c', 'cC'],
          ['d', 'dD'],
          ['e', 'eE'],
        ]);
      el.children = {
        a: el.children.a,
        b: el.children.c,
        c: el.children.b,
        d: HTML.li(listeners, 'f'),
        e: el.children.e,
      };
      el.children.e = HTML.li(listeners, 'g');
      assert.deepStrictEqual(
        Object.entries(el.children).map(([k, v]) => [k, v.children]),
        [
          ['a', 'aA'],
          ['b', 'cC'],
          ['c', 'bB'],
          ['d', 'fF'],
          ['e', 'gG'],
        ]);
      assert.deepStrictEqual(
        [...Object.values(el.children)].map(v => v.element),
        [...el.element.children]);
    });

    it('shadow', function () {
      assert(Shadow('section', [HTML.p()]).element.outerHTML === '<section></section>');
      assert(Shadow.section([HTML.p()]).element.outerHTML === '<section></section>');
      assert(Shadow.section([HTML.p()]).element.shadowRoot!.innerHTML === '<p></p>');
      assert(Shadow.section([HTML.p()]).children[0].element.outerHTML === '<p></p>');
      assert(Shadow.section((h, t) => h(t, [html('p')])).element.shadowRoot!.innerHTML === '<p></p>');
      assert(Shadow.section((h, t) => shadow(h(t, [html('p')])).host as HTMLElement).element.shadowRoot!.innerHTML === '<p></p>');
      assert(Shadow.section((h, t) => shadow(h(t, [html('p')]), { mode: 'closed' }).host as HTMLElement).element.shadowRoot === null);
      assert(Shadow.section([HTML.p()], (h, t) => shadow(h(t), { mode: 'closed' }).host as HTMLElement).element.shadowRoot === null);
      assert(Shadow.section([HTML.p()], (h, t) => shadow(h(t), { mode: 'closed' }).host as HTMLElement).children[0].element.outerHTML === '<p></p>');
    });

  });

  describe('usage', function () {
    it('component', function () {
      class Component implements El {
        private readonly dom = HTML.section({
          style: HTML.style(`$scope ul { width: 100px; }`),
          content: HTML.ul([
            HTML.li(`item`)
          ]),
        });
        public readonly element = this.dom.element;
        public get children() {
          return this.dom.children.content.children;
        }
        public set children(children) {
          this.dom.children.content.children = children;
        }
      }

      const comp = new Component();
      assert(comp.children[0].children === 'item');
      comp.children = [
        HTML.li('Item')
      ];
      assert(comp.children[0].children === 'Item');
      assert(HTML.div([comp]));
    });

    it('component shadow', function () {
      class Component implements El {
        private readonly dom = Shadow.section({
          style: HTML.style(`ul { width: 100px; }`),
          content: HTML.ul([
            HTML.li(`item`)
          ]),
        });
        public readonly element = this.dom.element;
        public get children() {
          return this.dom.children.content.children;
        }
        public set children(children) {
          this.dom.children.content.children = children;
        }
      }

      const comp = new Component();
      assert(comp.children[0].children === 'item');
      comp.children = [
        HTML.li('Item')
      ];
      assert(comp.children[0].children === 'Item');
      assert(HTML.div([comp]));
    });

    it('component coroutine', function () {
      class Component extends Coroutine implements El {
        constructor() {
          super(async function* (this: Component) {
            assert(this.element);
            assert(this.children);
            this.children = this.children.map(el => {
              el.children = el.children.toUpperCase();
              return el;
            });
            while (true) {
              yield;
            }
          }, { trigger: 'element', capacity: 0 });
          assert(this.children[0].children === 'ITEM');
        }
        private readonly dom = Shadow.section({
          style: HTML.style(`ul { width: 100px; }`),
          content: HTML.ul([
            HTML.li(`item`)
          ]),
        });
        public readonly element = this.dom.element;
        public get children() {
          return this.dom.children.content.children;
        }
        public set children(children) {
          this.dom.children.content.children = children;
        }
      }

      const comp = new Component();
      assert(comp.children[0].children === 'ITEM');
      comp.children = [
        HTML.li('item')
      ];
      assert(comp.children[0].children === 'item');
      assert(HTML.div([comp]));
    });

    it('translate', function () {
      const i18n = i18next.createInstance({
        lng: 'en',
        resources: {
          en: {
            translation: {
              "a": "{{data}}",
            }
          }
        }
      });
      interface TransDataMap {
        'a': { data: string; };
      }
      const Trans = API<HTMLElementTagNameMap>((
        tag: keyof HTMLElementTagNameMap,
        _?: Attrs | Children,
        children?: keyof TransDataMap,
        data?: TransDataMap[keyof TransDataMap],
      ) =>
        html(tag, {
          onmutate: children
            ? ev =>
                i18n.init((err, t) =>
                  (ev.target as HTMLElement).textContent = err
                    ? 'Failed to init i18next.'
                    : t(children, data))
            : void 0,
        }));
      const data = <K extends keyof TransDataMap>(data: TransDataMap[K]) =>
        <T extends string, E extends Element>(
          factory: (tag: T, attrs: Attrs, children: string, data: TransDataMap[keyof TransDataMap]) => E,
          tag: T,
          attrs: Attrs,
          children: K,
        ): E =>
          factory(tag, attrs, children, data);

      const el = Trans.span('a', data({ data: 'A' }));
      assert(el.children === 'A');
      assert(el.element.textContent === 'A');
    });

  });

});
